
/*
  cc -g -o saskme_frontend saskme_frontend.c
*/

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "saskme_frontend.h"

/*

 This is an example frontend for program bin/saskme resp. command 
 scdbackup_askme. It shows how to start the program, how to submit input
 and how to receive output.

 The program creates two pipes and forks a child process which then
 transforms itself into a scdbackup_askme run which gets forwarded the
 arguments of the frontend program together with some fixed settings.
 The pipes replace stdin and stdout of the child process. 
 One of the fixed arguments is "-pkt_output" which consolidates stdin and
 stdout of scdbackup_askme on stdout and marks the output packets so that
 they can be distinguished on the receiver side. 
 This is done in function  Saskme_side() .
  
 The original process enters function Frontend_side() which uses the pipes
 to send saskme options to the child process and to read the replies.
 The received output lines get distinguished by their -pkt_output channel ids.
 Frontend_handle_reply() reads them line by line, prints them to stdout and
 ends as soon as the output of the submitted option is complete.  

 The saskme-specific activities are performed by class Saskme_frontenD.
 The application submits plain option lines (including newline) via
 Saskme_frontend_send() and receives plain output lines together with
 the channel id via Saskme_frontend_rcv(). The return value of 
 Saskme_frontend_rcv() indicates wether the transaction is completed.

 Preceeding each option line from the application, Saskme_frontend_send()
 submits to the child a -mark option:
   -mark:-input-number-#
 # is a counter which is incremented with each input line. This mark will
 surround the child output caused by the input line.
 Function Saskme_frontend_rcv() will consider the output complete when the
 mark is received for the second time. First time is from -mark itself,
 second time is at the end of the option's output.

 saskme options which interfere with the functionality of this fork driven
 frontend are blocked within  Saskme_side()  and  Frontend_prompt_input() .

 Frontend_prompt_input() otherwise does not much refer to saskme.
 Frontend_handle_reply()  demonstrates a saskme reply loop and the
 distinction of channel ids. It's activities are a mere dummy, though.

*/


int Saskme_side();
int Frontend_side();


int main(argc,argv)
int argc;
char **argv;
{
 int ret;
 int stdin_pipe[2],stdout_pipe[2];
 pid_t pid;

 if(argc<=1) {
   fprintf(stderr,"usage: %s infosource [options|searchtexts]\n",argv[0]);
   exit(1);
 }

 /* create pipes */
 if(pipe(stdin_pipe)==-1) {
cannot_make_pipe:;
   fprintf(stderr,"%s : cannot create two pipes\n   reason given:   %s\n",
                  argv[0],strerror(errno));
   exit(2);
 }
 if(pipe(stdout_pipe)==-1)
   goto cannot_make_pipe;


 /* fork and execute */
 pid= fork();
 if(pid==-1) {
   fprintf(stderr,"%s : cannot fork a new process\n   reason given:   %s\n",
                  argv[0],strerror(errno));
   exit(3);

 } else if(pid>0) { /* original process */
   ret= Frontend_side(argv[0],stdin_pipe,stdout_pipe,0);
   if(ret<=0)
     exit(4);
 } else { /* child process */
   Saskme_side(argc,argv,stdin_pipe,stdout_pipe,0);
   exit(5);
 } 
 exit(0);
}


/* ----------------------------------------------------------------------- */


int Saskme_side(argc,argv,stdin_pipe,stdout_pipe,flag)
int argc;
char **argv;
int stdin_pipe[2],stdout_pipe[2];
int flag;
{
 int i,idx,ret,reply_complete;
 char **exec_argv;
/*
 static char child_prog[]= { "/home/thomas/cd_backup_planer_dir/bin/saskme" };
*/
 static char child_prog[]= { "scdbackup_askme" };
 static char pre_fixed_argv[][80]= {
                  "-pkt_output",
                  "-use_stdin",
                  "-result_format:levhdr:volno:target_esc:source" 
                                   };
 static int pre_fixed_argc= 3;
 static char post_fixed_argv[][80]= {
                  "-page:0",
                  "-mark:-saskme-child-started-up",
                  "-mark:" 
                                   };
 static int post_fixed_argc= 3;


 /* Prepare argument array */
 exec_argv= (char **) malloc((argc+1+pre_fixed_argc+post_fixed_argc)
                             *sizeof(char *));
 if(exec_argv==NULL) {
   fprintf(stderr,"%s : cannot allocate %ld bytes for argument pointer arry\n",
                  argv[0],(argc+1)*sizeof(char *));
   return(-1);
 }
 exec_argv[0]= "saskme";
 exec_argv[1]= argv[1];
 idx= 2;
 for(i=0;i<pre_fixed_argc;i++)
   exec_argv[idx++]= pre_fixed_argv[i];
 for(i= 2;i<argc;i++) {
   if(Saskme_frontend__forbidden_option(argv[i],0)) {
       fprintf(stderr,
               "\n== Option -page would confuse this program. Rejected.\n\n");
   } else
     exec_argv[idx++]= argv[i];
 }
 for(i=0;i<post_fixed_argc;i++)
   exec_argv[idx++]= post_fixed_argv[i];
 exec_argv[idx]= NULL;

 /* Connect stdin and stdout to the appropriate ends of the pipes. */
 close(0);
 ret= dup(stdin_pipe[0]);
 if(ret==-1) {
cannot_dup_pipe:;
   fprintf(stderr,
         "%s : cannot redirect standard i/o to pipes\n   reason given:   %s\n",
           argv[0],strerror(errno));
   free((char *) exec_argv);
   return(-1);
 }  
 close(1);
 ret= dup(stdout_pipe[1]);
 if(ret==-1)
   goto cannot_dup_pipe;

 /* close pipes so we register EOF by the father process */
 close(stdin_pipe[0]);
 close(stdin_pipe[1]);
 close(stdout_pipe[0]);
 close(stdout_pipe[1]);

 /* switch execution to the saskme program */
 execvp(child_prog,exec_argv);
 fprintf(stderr,"%s : cannot start program %s\n   reason given:   %s\n",
                argv[0],child_prog,strerror(errno));
 free((char *) exec_argv);
 return(-2);
}


/* ----------------------------------------------------------------------- */


int Frontend_prompt_input();
int Frontend_handle_reply();


int Frontend_side(prog_name,stdin_pipe,stdout_pipe,flag)
char *prog_name;
int stdin_pipe[2],stdout_pipe[2];
int flag;
{
 int ret,l;
 char *bufpt;
 struct Saskme_frontenD *saskme= NULL;

 ret= Saskme_frontend_new(&saskme,prog_name,stdin_pipe,stdout_pipe,0);
 if(ret<=0)
   goto ex;

 /* dialog loop */
 while(1) {

   if(saskme->starting_up) { 
     /* do not send option but consume output up to startup mark */;
   } else {

     /* obtain an input line from the user */
     ret= Frontend_prompt_input(saskme,0);
     if(ret<0)
       goto ex;
     if(ret==0)
       {ret= 1; goto ex;}

     /* send to child */
     ret= Saskme_frontend_send(saskme,0);
     if(ret<=0)
       goto ex;

   }

   /* receive reply up to completion and put it out to the terminal */
   ret= Frontend_handle_reply(saskme,1);
   if(ret<=0)
     goto ex;
   if(ret==2) /* child output EOF */
 break;

 }

 ret= 1;
ex:;
 Saskme_frontend_destroy(&saskme,0);
 return(ret);
}


int Frontend_handle_reply(saskme,flag)
struct Saskme_frontenD *saskme;
int flag;
/*
 bit0= suppress saskme prompt message
*/
/*
 return:
 <=0  error
   1  ok, transaction completed
   2  end, EOF at child stdout
*/
{
 int ret,reply_complete= 0,channel_id,first_mark;
 char *payload;

 while(!reply_complete) {
     
   ret= Saskme_frontend_rcv(saskme,&payload,&channel_id,0);
   if(ret<0)
     goto ex; 
   if(ret==0) {
     fprintf(stderr,"%s : EOF at child stdout\n",saskme->prog_name);
     {ret= 2; goto ex;}
   }
   reply_complete= (ret==2);
   first_mark= (ret==3);

   /* distinguish channel ids */
   if(channel_id=='R') {
     printf("R: %s\n",payload);

   } else if(channel_id=='I') {
     /* a note or error message */
     if((flag&1) && (strcmp(payload,"enter option or search text :\n")==0 ||
                     strcmp(payload,"Option oder Suchtext eingeben :\n")==0))
       /* suppress saskme prompt in favor of own one */;
     else
       fprintf(stderr,"I: %s\n",payload);

   } else if(channel_id=='M') {
     if(!reply_complete && !first_mark) {
       fprintf(stderr,"[Frontend:] unexpected mark paket received: %s\n",
                      payload);
       fprintf(stderr,"[Frontend:] expected: %s\n",saskme->markpt);
     }
   } else {
     /* rogue stdout (from wherever) */
     printf("-: %s\n",payload);
   }

 }

 saskme->starting_up= 0;
 ret= 1;
ex:;
 return(ret);
}


int Frontend_prompt_input(saskme,flag)
struct Saskme_frontenD *saskme;
int flag;
{
 int ret,l,i;

once_again:;
 /* read input line from own stdin */
 fprintf(stderr,"[Frontend:] enter option or search text :\n");
 l= sizeof(saskme->input_line);
 if(fgets(saskme->input_line,l-1,stdin)==NULL) {
   fprintf(stderr,"%s : EOF at stdin\n",saskme->prog_name);
   return(0);
 }
 if(Saskme_frontend__forbidden_option(saskme->input_line,0))
   goto once_again;
 l= strlen(saskme->input_line);
 if(saskme->input_line[l-1]!='\n')
   strcat(saskme->input_line,"\n");
 return(1);
}


/* - - - - - - -  rip off here for the Saskme_frontenD class  - - - - - - - */


#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));


/* Note: the pipes are closed after the object created here is done */

int Saskme_frontend_new(frontend,prog_name,stdin_pipe,stdout_pipe,flag)
struct Saskme_frontenD **frontend;
char *prog_name;
int stdin_pipe[2],stdout_pipe[2];
int flag;
{
 struct Saskme_frontenD *o;

 *frontend= o= TSOB_FELD(struct Saskme_frontenD,1);
 if(o==NULL)
   return(-1);
 strcpy(o->prog_name,prog_name);
 o->stdin_pipe[0]= stdin_pipe[0];
 o->stdin_pipe[1]= stdin_pipe[1];
 o->stdout_pipe[0]= stdout_pipe[0];
 o->stdout_pipe[1]= stdout_pipe[1];
 o->fp_stdout_pipe= NULL;
 o->input_counter= 0;
 o->input_line[0]= 0;
 o->starting_up= 1;
 o->buffer[0]= 0; /* this is actually a binary buffer and not a string */
 o->payload= NULL;
 sprintf(o->mark_buffer,"-saskme-child-started-up\n");
 o->markpt= o->mark_buffer;
 o->mark_counter= 1;/* startup will only deliver one of these marks, not two */
 o->fp_stdout_pipe= fdopen(stdout_pipe[0],"r");
 if(o->fp_stdout_pipe==NULL) {
   fprintf(stderr,"Saskme_frontend_new: cannot fdopen stdout_pipe[0]\n");
   fprintf(stderr,"       reason given: %s\n",strerror(errno));
   goto failed;
 }
 /* close unused pipe ends so we register EOF by the child process */
 close(stdin_pipe[0]);
 close(stdout_pipe[1]);
 return(1);
failed:;
 Saskme_frontend_destroy(frontend,0);
 return(-1);
}


int Saskme_frontend_destroy(frontend,flag)
struct Saskme_frontenD **frontend;
int flag;
{
 if((*frontend)==NULL)
   return(0);
 if((*frontend)->fp_stdout_pipe)
   fclose((*frontend)->fp_stdout_pipe);
 else
   close((*frontend)->stdout_pipe[0]);
 close((*frontend)->stdin_pipe[1]);
 free((char *) (*frontend));
 *frontend= NULL;
 return(1);
}


int Saskme_frontend_send(o,flag)
struct Saskme_frontenD *o;
int flag;
{
 int ret;

 /* compose mark_buffer and send -mark option */;
 o->input_counter++;
 sprintf(o->mark_buffer,"\n-input-number-%ld\n",o->input_counter);
 o->markpt= o->mark_buffer+1;
 o->mark_counter= 0;
 ret= write(o->stdin_pipe[1],"-mark:",6);
 if(ret<=0) {
write_err_child:;
   fprintf(stderr,"%s : error during write to child process:\n   %s\n",
           o->prog_name,strerror(errno));
   {ret= 0; goto ex;}
 }
 ret= write(o->stdin_pipe[1],o->markpt,strlen(o->markpt));
 if(ret<=0)
   goto write_err_child;

 /* send input line */;
 ret= write(o->stdin_pipe[1],o->input_line,strlen(o->input_line));
 if(ret<=0)
   goto write_err_child;

 ret= 1;
ex:;
 return(ret);
}


int Saskme_frontend_rcv(o,received,channel_id,flag)
struct Saskme_frontenD *o;
char **received;                /* a pointer to an address within o->buffer */
int *channel_id;                /* will become one of : '-','I','R','M' */
int flag;
/*
 return:
 <0  error
  0  EOF at child stdout
  1  ok
  2  this is the end mark of a transaction
  3  this is the first of two marks of a transaction
*/
{
 int ret,l,mode,reply_complete= 0,first_mark= 0;
 char *bufpt,*payload;

 /* read packet from child output */
 if(fgets(o->buffer,sizeof(o->buffer)-1,o->fp_stdout_pipe)==NULL)
   {ret= 0; goto ex;}
 bufpt= o->buffer;

 /* interpret packet */
 payload= NULL;
 if(bufpt[0]!=0 && bufpt[1]==':') {

   /* obtain packet mode and eventually remove trailing newline */
   payload= strchr(bufpt+2,':');
   if(payload!=NULL) {
     *payload= 0;
     sscanf(bufpt+2,"%d",&mode);
     *payload= ':';
     payload++;
     if(*payload==' ')
       payload++;
   }
   if(!(mode&1)) {
     /* newline is not part of payload */
     l= strlen(payload);
     if(l>0)
       if(payload[l-1]=='\n')
         payload[l-1]= 0;
   }
 }

 /* distinguish channel ids */
 *channel_id= bufpt[0];
 if(payload!=NULL && bufpt[0]=='R') {
   /* a result line (or part of it) */;
 } else if(payload!=NULL && bufpt[0]=='I') {
   /* a note or error message */;
 } else if(payload!=NULL && bufpt[0]=='M') {
   /* a saskme -mark message */
   l= strlen(payload);
   if(l+1==strlen(o->markpt))
     if(strncmp(payload,o->markpt,l)==0) { /* we know: markpt ends with \n */
       o->mark_counter++;
       if(o->mark_counter>=2) {
         reply_complete= 1;
         o->starting_up= 0;
       } else
         first_mark= 1;
    }
 } else {
   /* rogue stdout (from wherever) */
   *channel_id= '-';
 }
 ret= 1+reply_complete+2*first_mark;
ex:
 if(payload==NULL) {
   payload= o->buffer;
   *channel_id= '-';
 }
 *received= payload;
 o->payload= payload;
 return(ret);
}


int Saskme_frontend__forbidden_option(line,flag)
char *line;
int flag;
{
 int i;
 static char forbidden_options[][20]= {
                        "-page","-mark","-pkt_output","-use_readline",""};

 for(i=0;forbidden_options[i][0]!=0;i++) {
   if(strncmp(line,forbidden_options[i],strlen(forbidden_options[i]))==0) {
     fprintf(stderr,
       "\n[Frontend:] Option %s would confuse this program. Rejected.\n\n",
       forbidden_options[i]);
     return(1);
   }
 }
 return(0);
}


